"""
This script outputs the url for pip to be used when logging in to aws codeartifact.

https://docs.aws.amazon.com/codeartifact/latest/ug/npm-auth.html

Here is an example response:
    $ aws codeartifact login --tool npm --repository "$ENV_NPM_REPOSITORY" --domain "$ENV_DOMAIN"
        --domain-owner "$ENV_DOMAIN_OWNER"
"""

import os
import os.path
import sys
import datetime
import boto3
from furl import furl
import pyapikey

from pyawskit.configs import ConfigAwsCodeartifactNpm

KEY = "codeartifact_npm_config"
USERNAME = "aws"
USE_ENV = False
WRITE_FILE = True


def handle_data(url: str, short_url: str, password: str, replace: bool) -> None:
    """ really handle the data """
    if USE_ENV:
        url = furl(url)
        url.username = USERNAME  # type: ignore
        url.password = password  # type: ignore
        url.path.add("simple/")  # type: ignore
        print(f"export CODEARTIFACT_AUTH_TOKEN={password}")
        print(f"export npm_config_registry=\"{url}\"")
        print(f"export npm_config_{url}:always_auth=true")
        print("export npm_config_always_auth=true")
        print(f"export npm_config_{url}:_authToken={password}")
        print(f"export npm_config__authToken=\"{password}\"")
    if WRITE_FILE:
        filename = os.path.join(os.path.dirname(os.path.dirname(__file__)), ".npmrc")
        if not os.path.isfile(filename) or replace:
            with open("/dev/tty", "wt", encoding="utf-8") as stream:
                print(f"Writing new file [{filename}]...", file=stream)
            dirname = os.path.dirname(filename)
            if not os.path.isdir(dirname):
                os.makedirs(dirname)
            with open(filename, "wt", encoding="utf-8") as stream:
                stream.write(f"{short_url}:_authToken={password}\n")
                stream.write(f"registry={url}\n")


def run() -> None:
    env_npm_domain = ConfigAwsCodeartifactNpm.env_npm_domain
    env_npm_domain_owner = ConfigAwsCodeartifactNpm.env_npm_domain_owner
    env_npm_repository = ConfigAwsCodeartifactNpm.env_npm_repository
    temp_store = pyapikey.core.TempStore()
    if temp_store.has(KEY):
        data = temp_store.get(KEY)
        expiration = data["expiration"]
        now = datetime.datetime.now()
        if expiration > now:
            d_url = data["url"]
            d_short_url = data["short_url"]
            d_password = data["password"]
            handle_data(url=d_url, short_url=d_short_url, password=d_password, replace=False)
            sys.exit()

    with open("/dev/tty", "wt", encoding="utf-8") as f:
        print("Creating new temp key for codeartifact npm", file=f)

    client = boto3.client('codeartifact')
    res = client.get_authorization_token(
        domain=env_npm_domain,
        domainOwner=env_npm_domain_owner,
        durationSeconds=43200,
    )
    d_password = res["authorizationToken"]
    expiration = res["expiration"]
    res = client.get_repository_endpoint(
        domain=env_npm_domain,
        domainOwner=env_npm_domain_owner,
        repository=env_npm_repository,
        format="npm",
    )
    d_url = res["repositoryEndpoint"]
    if d_url.startswith("https:"):
        d_short_url = d_url[6:]

    data = {}
    data["url"] = d_url
    data["short_url"] = d_short_url
    data["password"] = d_password
    data["expiration"] = expiration.replace(tzinfo=None)
    temp_store.set(KEY, data)
    temp_store.save()

    handle_data(url=d_url, short_url=d_short_url, password=d_password, replace=True)



import os.path
import shutil
import subprocess
import sys

from typing import Any, List

import boto3
# noinspection PyPackageRequirements
import botocore
# noinspection PyPackageRequirements
import botocore.exceptions
import pypipegzip


def copyfileobj(source, destination, buffer_size=1024 * 1024):
    """
    Copy a file from source to destination. source and destination
    must be file-like objects, i.e. any object with a read or
    write method, like for example StringIO.
    """
    while True:
        copy_buffer = source.read(buffer_size)
        if not copy_buffer:
            break
        destination.write(copy_buffer)


def gzip_file(file_in: str, file_out: str) -> None:
    with open(file_in, 'rb') as f_in:
        f_out = pypipegzip.pypipegzip.zipopen(file_out, 'wb')
        shutil.copyfileobj(f_in, f_out)
        f_out.close()


def gzip_file_process(file_in: str, file_out: str) -> None:
    subprocess.check_call(
        f"gzip < {file_in} > {file_out}",
        shell=True,
    )


def object_exists(s3_connection, check_bucket_name: str, object_name: str) -> bool:
    """
    It seems funny we have to write this method but it seems that this
    is the way to check if an s3 bucket has an object.

    http://stackoverflow.com/questions/33842944/check-if-a-key-exists-in-a-bucket-in-s3-using-boto3

    :param s3_connection:
    :param check_bucket_name:
    :param object_name:
    :return: whether the object exists
    """
    try:
        # s3_connection.Object(check_bucket_name, object_name).load()
        s3_connection.head_object(Bucket=check_bucket_name, Key=object_name)
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == "404":
            return False
        raise
    return True


def object_exists_bucket(bucket_obj, object_name: str) -> bool:
    """
    see the note above

    :param bucket_obj:
    :param object_name:
    :return: whether the object exists
    """
    try:
        bucket_obj.Object(object_name).load()
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == "404":
            return False
        raise
    return True


def catch_all(the_function):
    def new_function(*args, **kwargs):
        # pylint: disable=broad-except
        try:
            return the_function(*args, **kwargs)
        except Exception as e:
            print('got exception', e)
            sys.exit(1)
    return new_function


def compress_one_file(_list: List[Any]) -> Any:
    pass


def process_one_file(basename, full_name, compressed_basename, full_compressed_name, bucket_name):
    s3 = boto3.resource('s3')
    bucket = s3.Bucket(bucket_name)

    print('downloading', full_name, basename)
    bucket.download_file(full_name, basename)

    print('zipping', basename, compressed_basename)
    gzip_file_process(basename, compressed_basename)

    print('upload', compressed_basename, full_compressed_name)
    with open(compressed_basename, 'rb') as file_handle:
        new_object_summary = s3.ObjectSummary(bucket_name, full_compressed_name)
        new_object_summary.put(Body=file_handle)
        del new_object_summary

    print('removing', basename, compressed_basename)
    os.unlink(basename)
    os.unlink(compressed_basename)


def print_exception(e):
    print('exception happened', e)
    sys.exit(1)



import errno
import json
import logging
import os
import socket
import stat
import subprocess
import sys
import time
from typing import List, Optional

import boto3
import requests


def load_json_config(
    name: str,
):
    home = os.getenv("HOME")
    assert home is not None
    path = os.path.join(
        home,
        ".pyawskit",
        name + ".json",
    )
    with open(path, "rt") as input_handle:
        obj = json.load(input_handle)
    return obj


def wait_net_service(
    server: str,
    port: int,
    timeout: int = 600,
) -> bool:
    """
        Wait for network service to appear
        http://code.activestate.com/recipes/576655-wait-for-network-service-to-appear/
        :param port:
        :param server:
        :param timeout
    """
    s = socket.socket()
    end: float
    if timeout:
        # time module is needed to calc timeout shared between two exceptions
        end = time.time() + timeout

    while True:
        try:
            if timeout:
                next_timeout = end - time.time()
                if next_timeout < 0:
                    return False
                s.settimeout(next_timeout)

            s.connect((server, port))

        except socket.timeout:
            # this exception occurs only if timeout is set
            if timeout:
                return False

        except socket.error as err:
            # catch timeout exception from underlying network library
            # this one is different from socket.timeout
            if err.errno != errno.ETIMEDOUT:
                raise
        else:
            s.close()
            return True


def get_disks() -> List[str]:
    # another option is to take the output of lsblk(1)
    # which is not amazon specific and has nice format of data.
    url = "http://169.254.169.254/latest/meta-data/block-device-mapping"
    r = requests.get(url, timeout=5)
    disks = []
    for line in r.content.decode().split("\n"):
        if not line.startswith('ephemeral'):
            continue
        ephemeral_url = f"http://169.254.169.254/latest/meta-data/block-device-mapping/{line}"
        r = requests.get(ephemeral_url, timeout=5)
        assert r.status_code == 200
        content = r.content.decode()
        assert content.startswith('sd')
        assert len(content) == 3
        letter = content[2]
        device = f"/dev/xvd{letter}"
        mode = os.stat(device).st_mode
        if not stat.S_ISBLK(mode):
            continue
        disks.append(device)
    return disks


def erase_partition_table(disk: str) -> None:
    logger = logging.getLogger(__name__)
    logger.info("erasing partition table on disk [%s]", disk)
    subprocess.check_call([
        "/bin/dd",
        "if=/dev/zero",
        f"of={disk}"
        "bs=4096",
        "count=1024",
    ], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)


def reread_partition_table() -> None:
    logger = logging.getLogger(__name__)
    logger.info("making OS re-read partition tables...")
    subprocess.check_call([
        "/sbin/partprobe",
    ])


def all_regions() -> List[str]:
    client = boto3.client('ec2')
    regions = [region['RegionName'] for region in client.describe_regions()['Regions']]
    return regions


def format_device(disk: str, label: Optional[str] = None) -> None:
    logger = logging.getLogger(__name__)
    logger.info(f"formatting the new device [{disk}]")
    args = [
        "/sbin/mkfs.ext4",
        disk,
    ]
    if label is not None:
        args.extend([
            "-L",
            label,
        ])
    subprocess.check_call(args, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)


def update_file(
    filename: str,
    pattern: str,
    do_all: bool = False,
):
    logger = logging.getLogger(__name__)

    config_file = os.path.expanduser(filename)
    if os.path.isfile(config_file):
        with open(config_file) as file_handle:
            lines = file_handle.readlines()
    else:
        lines = []

    # cut down auto generated lines if they exist...
    comment_line = "# DO NOT EDIT BEYOND THIS LINE - LEAVE THIS LINE AS IS\n"
    if comment_line in lines:
        location_of_comment_line = lines.index(comment_line)
        lines = lines[:location_of_comment_line]

    filter_file = os.path.expanduser("~/.pyawskit/aws_filter.json")
    if os.path.isfile(filter_file):
        logger.info(f"reading [{filter_file}]...")
        with open(filter_file) as file_handle:
            filters = json.loads(file_handle.read())
    else:
        logger.info(f"no filter file [{filter_file}] exists...")
        filters = []

    # look for servers matching the query, in all regions
    instances = []
    for region in all_regions():
        ec2 = boto3.resource('ec2', region_name=region)
        if do_all:
            instances.extend(list(ec2.instances.all()))
        else:
            instances.extend(list(ec2.instances.filter(Filters=filters)))
    num_of_instances = len(instances)
    logger.info(f"Found {num_of_instances} instance(s)")
    # assert num_of_instances > 0

    # add the comment line
    lines.append(comment_line)

    # add bunch of lines for each server
    added = 0
    for instance in instances:
        if instance.tags is None:
            continue
        tags_dict = {}
        for tag in instance.tags:
            tags_dict[tag["Key"]] = tag["Value"]
        if "Name" not in tags_dict:
            continue
        host = tags_dict["Name"]
        if host == "" or " " in host:
            logger.info(f"Name [{host}] for host is bad. Try non empty names without spaces...")
            continue
        # public_ip = instance.public_dns_name
        # if public_ip == "":
        #    continue
        # logger.info(dir(instance))
        private_ip = instance.private_ip_address
        if private_ip == "":
            continue
        pattern_to_add = pattern.format(
            host=host,
            ip=private_ip,
            key_name=instance.key_name,
        )
        lines.extend(pattern_to_add)
        added += 1

    # print the final lines to the config file
    with open(config_file, "wt") as file_handle:
        file_handle.writelines(lines)
    logger.info(f"Added {added} instance(s)")
    logger.info(f"Written {config_file}")


def mount_disk(disk: str, folder: str) -> None:
    logger = logging.getLogger(__name__)
    logger.info("mounting the new device [%s, %s]", disk, folder)
    if not os.path.isdir(folder):
        os.makedirs(folder)
    subprocess.check_call([
        "/bin/mount",
        disk,
        folder,
    ])


ssh_config_pattern = """Host {host}
\tHostName {ip}
\tIdentityFile ~/.pyawskit/keys/{key_name}.pem
\tIdentitiesOnly yes
\tUser ubuntu
\tStrictHostKeyChecking no
\tLogLevel ERROR
\tCheckHostIP no
\t# UserKnownHostsFile /dev/null
"""

etc_hosts_pattern = "{ip} {host}\n"


def update_ssh_config(all_hosts: bool):
    update_file(filename="~/.ssh/config.d/99_dynamic.conf", pattern=ssh_config_pattern, do_all=all_hosts)


def update_etc_hosts(all_hosts: bool):
    update_file(filename="/etc/hosts", pattern=etc_hosts_pattern, do_all=all_hosts)


def run_devnull(args: List[str]) -> None:
    subprocess.check_call(
        args,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL
    )


def check_root() -> None:
    """
    check that we are running as root
    """
    if not os.geteuid() == 0:
        sys.exit('script must be run as root')


def touch(filename: str) -> None:
    with open(filename, "w"):
        pass


def do_hush_login():
    filename = os.path.expanduser("~/.hushlogin")
    touch(filename)



import logging
from time import sleep
from typing import List, Callable, Dict

from pyawskit.common import load_json_config, wait_net_service

# http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/spot-requests.html
aws_spot_instance_types = {
    "c3.8xlarge",
    "c4.8xlarge",
    "d2.8xlarge",
    "g2.8xlarge",
    "i2.8xlarge",
    "m4.10xlarge",
    "m4.16xlarge",
    "p2.16xlarge",
    "r3.8xlarge",
    "r4.16xlarge",
    "x1.32xlarge",
}


def log_func_name(func: Callable):

    def inner(*kw, **tw):
        logger = logging.getLogger(__name__)
        logger.info(func.__name__)
        return func(*kw, **tw)
    return inner


class ProcessData:
    def __init__(self, name: str):
        self.p_name: str = name
        self.p_launch_config = load_json_config("launch_config")
        self.p_config = self.p_launch_config[self.p_name]
        self.p_price = float(self.p_config["price"])
        self.p_count = int(self.p_config["count"])
        self.p_region = self.p_config["region"]
        self.p_dry_run: bool = False
        self.p_type: str = "one-time"

    def get_price(self):
        return self.p_config["price"]


@log_func_name
def request_spot_instances(client, pd: ProcessData):
    r_request_spot_instances = client.request_spot_instances(
        DryRun=pd.p_dry_run,
        SpotPrice=str(pd.p_price),
        InstanceCount=pd.p_count,
        LaunchSpecification=pd.p_config["launch_config"],
        Type=pd.p_type,
    )
    return r_request_spot_instances


@log_func_name
def wait_using_waiter(client, pd: ProcessData, request_ids: List[str]):
    waiter = client.get_waiter('spot_instance_request_fulfilled')
    ret = waiter.wait(
        DryRun=pd.p_dry_run,
        SpotInstanceRequestIds=request_ids,
    )
    return ret


@log_func_name
def poll_requests_till_done(client, pd: ProcessData, request_ids: List[str]):
    response = client.describe_spot_instance_requests(
        DryRun=pd.p_dry_run,
        SpotInstanceRequestIds=request_ids,
    )
    return response


@log_func_name
def poll_instances_till_done(ec2, pd: ProcessData, request_ids: List[str]):
    instances = ec2.instances.filter(Filters=[{'Name': 'spot-instance-request-id', 'Values': request_ids}])
    while len(list(instances)) < pd.p_count:
        sleep(1)
        instances = ec2.instances.filter(Filters=[{'Name': 'spot-instance-request-id', 'Values': request_ids}])
    return instances


@log_func_name
def tag_resources(client, resource_ids: List[str], tags: object):
    response = client.create_tags(
        Resources=resource_ids,
        Tags=tags,
    )
    return response


@log_func_name
def attach_disks(ec2, instance_ids: List[str], disks: List[Dict[str, str]]):
    """
    This function attaches a list of disks to the instances in the list
    :param ec2:
    :param instance_ids:
    :param disks:
    :return:
    """
    responses = []
    for p_instance_id in instance_ids:
        for disk in disks:
            response = attach_disk(
                ec2=ec2,
                instance_id=p_instance_id,
                volume_id=disk['volume_id'],
                device=disk['device'],
            )
            responses.append(response)
    return responses


@log_func_name
def wait_for_ssh(instances):
    for instance in instances:
        wait_net_service(
            server=instance.private_ip_address,
            port=22,
        )


@log_func_name
def attach_disk(ec2, instance_id: str, volume_id: str, device: str):
    volume = ec2.Volume(volume_id)
    response = volume.attach_to_instance(
        InstanceId=instance_id,
        Device=device,
    )
    return response



import setuptools


def get_readme():
    with open('README.rst') as f:
        return f.read()


setuptools.setup(
    # the first three fields are a must according to the documentation
    name="pyawskit",
    version="0.1.74",
    packages=[
        "pyawskit",
    ],
    # from here all is optional
    description="Pyawskit is your AWS Swiss Army Knife",
    long_description=get_readme(),
    long_description_content_type="text/x-rst",
    author="Mark Veltzer",
    author_email="mark.veltzer@gmail.com",
    maintainer="Mark Veltzer",
    maintainer_email="mark.veltzer@gmail.com",
    keywords=[
        "aws",
        "utils",
        "python",
        "API",
        "command",
        "line",
    ],
    url="https://veltzer.github.io/pyawskit",
    download_url="https://github.com/veltzer/pyawskit",
    license="MIT",
    platforms=[
        "python3",
    ],
    install_requires=[
        "pylogconf",
        "pyfakeuse",
        "pypipegzip",
        "tqdm",
        "requests",
        "boto3",
        "pymount",
        "ujson",
        "sultan",
        "pytconf",
        "pyapikey",
        "furl",
        "docker-py",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
    ],
    entry_points={"console_scripts": [
        "pyawskit=pyawskit.main:main",
    ]},
)



from pytconf import Config, ParamCreator


class ConfigName(Config):
    """
    Parameters to select the name of the config to launch
    """
    name = ParamCreator.create_str(help_string="index to use")


class ConfigFilter(Config):
    """
    Parameters to select if the use the filter or not
    """
    filter = ParamCreator.create_bool(
        default=True,
        help_string="use filter for the instances",
    )


class ConfigWork(Config):
    """
    Parameters to control the actualy device work
    """
    device_file = ParamCreator.create_str(
        default="/dev/md0",
        help_string="device to use",
    )
    mount_point = ParamCreator.create_str(default="/mnt/raid0", help_string="mount point")
    mdadm_config_file = ParamCreator.create_str(default='/etc/mdadm/mdadm.conf', help_string="mdadm config file")
    mdadm_binary = ParamCreator.create_str(default="/sbin/mdadm", help_string="mdam tool")
    fstab_filename = ParamCreator.create_str(default="/etc/fstab", help_string="fstab filename")
    file_system_type = ParamCreator.create_str(default="ext4", help_string="file system type")
    name_of_raid_device = ParamCreator.create_str(default="MY_RAID", help_string="name of raid device")
    start_stop_queue = ParamCreator.create_bool(default=False, help_string="do start stop queue")
    write_etc_mdadm = ParamCreator.create_bool(default=False, help_string="write /etc/mdadm")
    add_line_to_fstab = ParamCreator.create_bool(default=False, help_string="add line to /etc/fstab")


class ConfigAwsCodeartifactNpm(Config):
    env_npm_domain = ParamCreator.create_str(help_string="env npm domain")
    env_npm_domain_owner = ParamCreator.create_str(help_string="env npm domain owner")
    env_npm_repository = ParamCreator.create_str(help_string="env npm repository")


class ConfigAwsCodeartifactPip(Config):
    env_pip_domain = ParamCreator.create_str(help_string="env pip domain")
    env_pip_domain_owner = ParamCreator.create_str(help_string="env pip domain owner")
    env_pip_repository = ParamCreator.create_str(help_string="env pip repository")



import sys
import datetime
import base64

import requests.exceptions
import docker
import boto3
import pyapikey

KEY = "ecr_password"


def handle_data(user: str, password: str, proxyEndpoint: str) -> None:
    """ really do something with the data """
    client = docker.Client()
    try:
        client.login(
            username=user,
            password=password,
            registry=proxyEndpoint,
        )
    except requests.exceptions.ConnectionError:
        print(f"ConnectionError for {user} {password} {proxyEndpoint}", file=sys.stderr)
        sys.exit(1)


def run() -> None:
    temp_store = pyapikey.core.TempStore()
    if temp_store.has(KEY):
        data = temp_store.get(KEY)
        expiration = data["expiration"]
        now = datetime.datetime.now()
        if expiration > now:
            d_user = data["user"]
            d_password = data["password"]
            d_proxyEndpoint = data["proxyEndpoint"]
            handle_data(
                user=d_user,
                password=d_password,
                proxyEndpoint=d_proxyEndpoint,
            )
            sys.exit()

    with open("/dev/tty", "wt", encoding="utf-8") as f:
        print("Creating new temp key for ecr", file=f)

    client = boto3.client("ecr")
    res = client.get_authorization_token()
    authorizationData = res["authorizationData"][0]
    d_authorizationToken = authorizationData["authorizationToken"]
    d_expiresAt = authorizationData["expiresAt"]
    d_proxyEndpoint = authorizationData["proxyEndpoint"]

    # decode the password
    base64decode = base64.b64decode(d_authorizationToken).decode("ASCII")
    (d_user, d_password) = base64decode.split(":")
    handle_data(
        user=d_user,
        password=d_password,
        proxyEndpoint=d_proxyEndpoint,
    )

    # save the token
    data = {}
    data["user"] = d_user
    data["password"] = d_password
    data["proxyEndpoint"] = d_proxyEndpoint
    data["expiration"] = d_expiresAt.replace(tzinfo=None)
    temp_store.set(KEY, data)
    temp_store.save()
